#!/usr/bin/env python
# -*- coding: utf-8 -*-

def mod(a, b):
    """
    Returns a % b, with a % 0 = a.

    Parameters
    ----------
    a : int
        The first argument in a % b.
    b : int
        The second argument in a % b.

    Returns
    -------
    int
        `a` modulus `b`.

    Examples
    ---------
    >>> mod(5, 2)
    1
    >>> mod(5, 0)
    5
    """
    if b == 0:
        return a
    return a % b




if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = True)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import types
from sympy import Matrix, diag, latex
from abelian.utils import mod
from abelian.groups import LCA
from abelian.linalg.utils import delete_zero_columns, nonzero_diag_as_list, \
    matrix_mod_vector, order_of_vector, remove_cols, remove_rows
from abelian.linalg.factorizations import smith_normal_form
from abelian.linalg.solvers import solve_epi


class HomLCA:
    """
    A homomorphism between LCAs.
    """

    # The types allowed as entries in A
    _A_entry_types = (int, float, complex)

    def __init__(self, A, target = None, source = None):
        """
        Initialize a homomorphism.

        Parameters
        ----------
        A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
            A sympy matrix representing the homomorphism. The user may also
            use a list of lists in the form [row1, row2, ...] as input.
        target : LCA or list
            The target of the homomorphism. If None, a discrete target of
            infinite period is used as the default.
        source : LCA or list
            The source of the homomorphism. If None, a discrete source of
            infinite period is used as the default.

        Examples
        ---------
        >>> phi = HomLCA([[1,2],
        ...               [3,4]])
        """

        A, target, source = self._verify_init(A, target, source)
        assert isinstance(target, LCA)
        assert isinstance(source, LCA)
        self.A = A
        self.target = target
        self.source = source

    @staticmethod
    def _verify_init(A, target, source):
        """
        Verify the inputs.

        Parameters
        ----------
        A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
            A sympy matrix representing the homomorphism. The user may also
            use a list of lists in the form [row1, row2, ...] as input.
        target : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`,
        list or LCA
            The target of the homomorphism. If None, a discrete target of
            infinite period is used as the default.
        source : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`,
        list or LCA
            The source of the homomorphism. If None, a discrete source of
            infinite period is used as the default.

        Returns
        -------
        A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        target : LCA
        source : LCA
        """

        # If the input matrix is a list of lists, convert to matrix
        if isinstance(A, list):
            A = Matrix(A)

        # If no target is given, assume free-to-free morphism
        if (target is None):
            target = [0] * A.rows
            source = [0] * A.cols

        # If target is given, and no source, assume left-free morphism
        elif (source is None):
            source = [0] * A.cols

        # If lists of matrices were passed, convert to LCAs
        if isinstance(target, (list, Matrix)):
            target = LCA(list(target))

        if isinstance(source, (list, Matrix)):
            source = LCA(list(source))

        return A, target, source


    def to_latex(self):
        """
        Write to latex string.

        Returns
        -------

        """
        latex_code = latex(self.A)
        latex_code = latex_code.replace(r'\left[\begin{matrix}',
                                        r'\begin{pmatrix}')
        latex_code = latex_code.replace(r'\end{matrix}\right]',
                                        r'\end{pmatrix}')
        format_args = self.source.to_latex(), self.target.to_latex()
        latex_code += r':{} \to {}'.format(*format_args)
        return latex_code


    def __getitem__(self, args):
        """
        Override the slice (`obj[a:b]`) operator.
        """
        if len(args) != 2:
            raise ValueError('__getitem__ takes 2 arguments.')

        slice1, slice2 = args
        new_A = self.A[slice1, slice2]
        new_target = self.target[slice1]
        new_source = self.source[slice2]
        return type(self)(new_A, new_target, new_source)


    def add(self, other):
        """
        Elementwise addition.

        Parameters
        ----------
        other

        Returns
        -------

        """
        # If the `other` argument is a numeric type
        if isinstance(other, self._A_entry_types):
            m, n = self.A.shape
            new_A = self.A + Matrix(m, n, lambda i,j : other)
            return type(self)(new_A, target=self.target, source=self.source)

        if isinstance(other, type(self)):
            if not self.target == other.target and self.source == other.source:
                raise ValueError('Sources and targets must match..')

            new_A = self.A + other.A
            return type(self)(new_A, target=self.target, source=self.source)

        format_args = type(self), type(other)
        raise ValueError('Cannot add {} and {}'.format(*format_args))

    def __add__(self, other):
        """
        Override the addition (`+`) operator.
        """
        return self.add(other)

    def __radd__(self, other):
        """
        Override the addition (`+`) operator.
        """
        return self.add(other)


    def pow(self, power):
        """
        Repeated composition.

        Parameters
        ----------
        power

        Returns
        -------

        """
        product = self.copy()
        for prod in range(power - 1):
            product *= self
        return product


    def __pow__(self, power, modulo=None):
        """
        Override the power (`**`) operator.
        """
        return self.pow(power)

    def copy(self):
        """
        Return a copy.
        """
        return type(self)(self.A, target = self.target, source = self.source)


    def compose(self, other):
        """
        Compose two homomorphisms.

        Parameters
        ----------
        other

        Returns
        -------
        >>> phi = HomFGA([[1, 0, 1],
        ...               [0, 1, 1]])
        >>> ker_phi = HomFGA([1, 1, -1])
        >>> (phi * ker_phi) == HomFGA([0, 0])
        True
        """
        # If the `other` argument is a numeric type
        if isinstance(other, self._A_entry_types):
            new_A = self.A * other
            return type(self)(new_A, target=self.target, source=self.source)

        if isinstance(other, type(self)):
            if not other.target == self.source:
                raise ValueError('Target of other must equal source of self.')

            new_A = self.A * other.A
            return type(self)(new_A, target=self.target, source=other.source)

        format_args = type(self), type(other)
        raise ValueError('Cannot compose/add {} and {}'.format(*format_args))

    def __mul__(self, other):
        """
        Override the multiplication (`*`) operator.
        """
        return self.compose(other)

    def __rmul__(self, other):
        """
        Override the multiplication (`*`) operator.
        """
        return self.__mul__(other)

    def stack_vert(self, other):
        """
        Stack vertically.

        Parameters
        ----------
        other

        Returns
        -------

        """
        pass

    def stack_horiz(self, other):
        """
        Stack horizontally.

        Parameters
        ----------
        other

        Returns
        -------

        """
        pass

    def stack_diag(self, other):
        """
        Stack diagonally.

        Parameters
        ----------
        other

        Returns
        -------

        """
        pass

    def evaluate(self, source_element):
        """
        Apply the morphism to an element.

        Parameters
        ----------
        source_element

        Returns
        -------

        """
        pass


    def to_HomFGA(self):
        """
        Convert object to HomFGA if possible.

        Returns
        -------

        """
        integer_entries = all([i % 1 == 0 for i in self.A])
        if self.source.is_FGA() and self.target.is_FGA() and integer_entries:
            return HomFGA(self.A, self.target, self.source)
        else:
            return self


    def equal(self, other):
        """
        Check equality.

        Parameters
        ----------
        other

        Returns
        -------

        """
        source_equal = self.source == other.source
        target_equal = self.target == other.target
        A_equal = self.A == other.A
        return source_equal and target_equal and A_equal

    def __eq__(self, other):
        """
        Override the equality (`==`) operator.
        """
        return self.equal(other)


    def __repr__(self):
        """
        Override the `repr()` function.
        """
        rep = 'source: {}     target: {}\n'.format(repr(self.source),
                                                   repr(self.target))
        rep += str(self.A)
        return rep



class HomFGA(HomLCA):
    """
    A homomorphism between two FGAs.
    """

    # The types allowed as entries in A
    _A_entry_types = (int,)

    def project_to_target(self):
        """
        Project columns to target group.

        Returns
        -------

        Examples
        --------
        >>> target = [7, 12]
        >>> phi = HomFGA([[15, 12],
        ...               [9,  17]], target = target)
        >>> phi_proj = HomFGA([[1, 5],
        ...                    [9, 5]], target = target)
        >>> phi.project_to_target() == phi_proj
        True
        """

        A = matrix_mod_vector(self.A, Matrix(self.target.periods))
        return type(self)(A, target = self.target, source = self.source)

    def project_to_source(self):
        """
        Compute periods.

        Returns
        -------

        Examples
        --------
        >>> target = [3, 6]
        >>> phi = HomFGA([[1, 0],
        ...               [3, 3]], target = target)
        >>> phi = phi.project_to_source()
        >>> phi.source.to_list() == [6, 2]
        True
        """

        # Find dimensions
        m, n = self.A.shape

        # Compute orders for all columns of A
        target_vect = Matrix(self.target.periods)
        source = [order_of_vector(self.A[:, i], target_vect) for i in range(n)]
        return type(self)(self.A, self.target, source)

    def kernel(self):
        """
        Compute the kernel homomorphism.

        Returns
        -------
        homomorphism : HomFGA
            The kernel homomorphism.

        Examples
        --------
        >>> phi = HomFGA([[1, 0, 1], [0, 1, 1]])
        >>> phi.kernel() == HomFGA([-1, -1, 1])
        True

        """
        # Horizontally stack A and ker(pi_2)
        A_ker_pi = self.A.row_join(delete_zero_columns(diag(*self.target)))

        # Compute SNF, get size and the kernel
        U, S, V = smith_normal_form(A_ker_pi)
        (m, n), r = self.A.shape, len(nonzero_diag_as_list(S))
        kernel_matrix = V[:n, r:]

        kernel = type(self)(kernel_matrix, target = self.source)
        return kernel.project_to_target()

    def cokernel(self):
        """
        Compute the cokernel homomorphism.

        Returns
        -------
        homomorphism : HomFGA
            The cokernel homomorphism.

        Examples
        --------
        >>> 1 == 1
        True

        """
        # Horizontally stack A and ker(pi_2)
        A_ker_pi = self.A.row_join(delete_zero_columns(diag(*self.target)))
        # Compute SNF, get size and the kernel
        U, S, V = smith_normal_form(A_ker_pi)
        diagonal = nonzero_diag_as_list(S)
        (m, n), r = self.A.shape, len(diagonal)
        quotient = diagonal + [0] * (m - r)

        # Initialize the cokernel morphism and project it onto the target
        coker = type(self)(U, target = quotient, source=self.target)
        return coker.project_to_target()

    def image(self):
        """
        Compute the image homomorphism.

        Returns
        -------
        homomorphism : HomFGA
            The image homomorphism.

        Examples
        --------
        >>> 1 == 1
        True

        """
        # Solve equation for the image
        coim = self.coimage()
        coim_target = Matrix(coim.target.periods)
        solved_mat = solve_epi(coim.A, self.A, coim_target)

        # Initialize morphism and return project onto target
        image = type(self)(solved_mat, self.target, source=coim.target)
        return image.project_to_target()

    def coimage(self):
        """
        Compute the coimage homomorphism.

        Returns
        -------
        homomorphism : HomFGA
            The coimage homomorphism.

        Examples
        --------
        >>> 1 == 1
        True

        """
        # Compute the cokernel of the kernel and return
        kernel = self.kernel()
        coimage = kernel.cokernel()
        return coimage

    def remove_trivial_groups(self):
        """
        Remove trivial from source and target.

        A group is trivial if it is discrete with period 1, i.e. Z_1.
        Removing trivial groups from the target group means removing the
        Z_1 groups from the target, along with the corresponding rows of
        the matrix representing the homomorphism.
        Removing trivial groups from the source group means removing the
        groups Z_1 from the source, i.e. removing every column (generator)
        with period 1.

        Returns
        -------
        homomorphism : HomFGA
            A homomorphism where the trivial groups have been removed from
            the source and the target. The corresponding rows and columns of
            the matrix representing the homomorphism are also removed.

        Examples
        --------
        >>> target = [1, 7]
        >>> phi = HomFGA([[2, 1], [7, 2]], target=target)
        >>> projected = HomFGA([[2]], target=[7], source = [7])
        >>> phi.project_to_source().remove_trivial_groups() == projected
        True

        """
        def trivial(period, discrete):
            return discrete and (period == 1)

        # Get indices where the value of the source is 1
        generator = enumerate(self.source._gen())
        cols_to_del = [i for (i, (d, p)) in generator if trivial(d, p)]
        new_A = remove_cols(self.A, cols_to_del)

        # Get indices where the value of the target is 1
        generator = enumerate(self.target._gen())
        rows_to_del = [i for (i, (d, p)) in generator if trivial(d, p)]
        new_A = remove_rows(new_A, rows_to_del)

        new_source = self.source.delete_by_index(cols_to_del)
        new_target = self.target.delete_by_index(rows_to_del)

        return type(self)(new_A, new_target, new_source)


def Homomorphism(A, target = None, source = None):
    """
    Initiliazes HomFGA or HomLCA depending on inputs.

    This factory function will initialize a HomFGA if both the source and
    target are FGAs, or None. If the source and targets are explicitly given
    as non-discrete LCAs then a HomLCA will be initialized.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
        A sympy matrix representing the homomorphism. The user may also
        use a list of lists in the form [row1, row2, ...] as input.
    target : LCA or list
        The target of the homomorphism. If None, a discrete target of
        infinite period is used as the default.
    source : LCA or list
        The source of the homomorphism. If None, a discrete source of
        infinite period is used as the default.

    Returns
    -------

    Examples
    --------
    >>> # If no source/target is given, the defalt is discrete (FGA)
    >>> phi = Homomorphism([1])
    >>> isinstance(phi, HomFGA)
    True
    >>> # If the target is continuous, a HomLCA instance will be returned
    >>> target = LCA(periods = [0], discrete = [False])
    >>> phi = Homomorphism([1], target = target)
    >>> isinstance(phi, HomLCA)
    True

    """
    A, target, source = HomLCA._verify_init(A, target, source)
    integer_entries = all([i % 1 == 0 for i in A])
    if target.is_FGA() and source.is_FGA() and integer_entries:
        return HomFGA(A, source, target)
    else:
        return HomLCA(A, source, target)





if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = True)


if __name__ == '__main__':

    print('Creating from lists')
    target = [7, 12]
    phi = HomFGA([[15, 12], [9, 17]], target=target)
    print(type(phi.source), type(phi.target))
    assert isinstance(phi.source, LCA)
    assert isinstance(phi.target, LCA)

    print('Creating from objects')
    target = LCA([7, 12])
    phi = HomFGA([[15, 12], [9, 17]], target=target)
    print(type(phi.source), type(phi.target))
    assert isinstance(phi.source, LCA)
    assert isinstance(phi.target, LCA)

    print('Creating from matrices')
    target = Matrix([7, 12])
    phi = HomFGA([[15, 12], [9, 17]], target=target)
    print(type(phi.source), type(phi.target))
    assert isinstance(phi.source, LCA)
    assert isinstance(phi.target, LCA)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sympy import Matrix
from abelian.utils import mod

class LCA(object):
    """
    Class for Locally Compact Abelian (LCA) groups.
    """


    _repr_dict = {True: 'd', False :'c'}
    _repr_dict_inv = {'d':True, 'c':False}

    def __init__(self, periods, discrete = None):
        """
        Initialize a new Locally Compact Abelian (LCA) group.

        Parameters
        ----------
        periods : list
            A list of periods, e.g. [6, 8, 11].
        discrete : list
            A list of booleans or ['d', 'c', ...] where 'd' stands for
            discrete and 'c' stands for continuous. If None, it defaults to
            discrete.


        Examples
        ---------
        >>> G1 = LCA([5, 6, 7])
        >>> G2 = LCA([5, 6, 7], [True, True, True])
        >>> G3 = LCA([5, 6, 7], ['d']*3)
        >>> (G1 == G2 == G3)
        True
        """

        periods, discrete = self._verify_init(periods, discrete)
        self.periods = periods
        self.discrete = discrete


    @classmethod
    def _verify_init(cls, periods, discrete):
        """
        Verify the user inputs.
        """

        if isinstance(periods, Matrix):
            periods = list(periods)

        if isinstance(discrete, Matrix):
            discrete = list(discrete)

        if len(periods) <= 0:
            raise ValueError('List of periods must have length >0.')

        if any(i < 0 for i in periods):
            raise ValueError('Every period must be >= 0.')

        if discrete is None:
            discrete = [True] * len(periods)

        if len(periods) != len(discrete):
            raise ValueError('Periods and list of discrete must match length.')

        for p, d in zip(periods, discrete):
            if not d and p not in [0, 1]:
                raise ValueError('Continuous groups must have period 0 or 1.')

        def map_to_bool(x):
            try:
                return cls._repr_dict_inv[x.lower()]
            except:
                return x

        discrete = [map_to_bool(i) for i in discrete]

        return periods, discrete


    def is_FGA(self):
        """
        Whether or not the LCA is a FGA.

        Returns
        -------
        order : bool
            True if the object is an FGA, False if not.

        Examples
        ---------
        >>> G = LCA([5, 1], [True, False])
        >>> G.is_FGA()
        False
        >>> G = LCA([1, 7], [False, True])
        >>> G.dual().is_FGA()
        True
        """
        return all(self.discrete)



    def __repr__(self):
        """
        Representation.
        """

        def repr_single(p, d):
            if p == 0:
                if d:
                    return r'Z'
                return r'R'
            if d:
                return r'Z_' + str(p) + ''
            return r'T_' + str(p) + ''

        joined = r', '.join(repr_single(p, d) for (p, d) in self._gen())
        return '[' + joined + ']'

        repr_dict = self._repr_dict
        return '[' + ', '.join([str(num) + '({})'.format(repr_dict[i]) for (
            num, i) in self._gen()]) + ']'

    def to_latex(self):
        """
        Write object to latex string.
        """
        def repr_single(p, d):
            if p == 0:
                if d:
                    return r'\mathbb{Z}'
                return r'\mathbb{R}'
            if d:
                return r'\mathbb{Z}_{' + str(p) + '}'
            return r'T_{' + str(p) + '}'

        return r' \oplus '.join(repr_single(p, d) for (p, d) in self._gen())

    def dual(self):
        """
        Return the Pontryagin dual.

        Returns
        -------
        group : LCA
            The Pontryagin dual of the LCA.

        Examples
        ---------
        >>> G = LCA([5, 1], [True, False])
        >>> H = LCA([5, 0], [True, True])
        >>> G.dual() == H
        True
        >>> G.dual().dual() == G
        True
        >>> self_dual = LCA([5])
        >>> self_dual.dual() == self_dual
        True
        """
        single_dual = self._dual_of_single_group
        dual_lists = list(single_dual(p, d) for (p, d) in self._gen())
        new_periods = [p for (p, d) in dual_lists]
        new_discrete = [d for (p, d) in dual_lists]
        return type(self)(periods = new_periods, discrete = new_discrete)

    def _gen(self):
        """
        Yield pairs of (period, discrete).
        """
        for period, discrete in zip(self.periods, self.discrete):
            yield (period, discrete)

    def project_element(self, element):
        """
        Project an element onto the group.

        Returns
        -------
        element : :py:class:`~sympy.matrices.dense.MutableDenseMatrix` or list
            The group element projected to the LCA.

        Examples
        ---------
        >>> from sympy import Matrix
        >>> G = LCA([5, 9])
        >>> g = [6, 22]
        >>> G.project_element(g)
        [1, 4]
        >>> g = Matrix([13, 13])
        >>> G.project_element(g) == Matrix([3, 4])
        True
        """
        generator = zip(element, self.periods)
        projected = [mod(element, period) for (element, period) in generator]

        # If the input is a list, return a list
        if isinstance(element, list):
            return projected

        # If the input is a sympy Matrix, return a sympy Matrix
        if isinstance(element, Matrix):
            return Matrix(projected)

    def copy(self):
        """
        Return a copy of the LCA.

        Returns
        --------
        group : LCA
            A copy of the object.


        Examples
        ---------
        >>> G = LCA([1, 5, 7], [False, True, True])
        >>> H = G.copy()
        >>> G == H
        True
        """
        return type(self)(periods = self.periods.copy(),
                          discrete = self.discrete.copy())

    def equal(self, other):
        """
        Whether or not two LCAs are equal.

        Parameters
        ----------
        other : LCA
            The LCA to compare with.

        Returns
        --------
        equal : bool
            Whether or not the LCAs are equal.


        Examples
        ---------
        >>> G = LCA([1, 5, 7], [False, True, True])
        >>> H = G.copy()
        >>> G == H  # The `==` operator is overloaded
        True
        >>> G.equal(H)  # Equality using the method
        True
        """
        return ((self.periods == other.periods) and
                (self.discrete ==  other.discrete))

    def __eq__(self, other):
        """
        Override the equality (`==`) operator.
        """
        return self.equal(other)

    def isomorphic(self):
        """
        TODO.
        """
        pass

    def direct_sum(self, other):
        """
        Returns the direct sum of two LCAs.

        Parameters
        ----------
        other : LCA
            The LCA to take direct sum with.

        Returns
        --------
        group : LCA
            The direct product of self and other.


        Examples
        ---------
        >>> G = LCA([5])
        >>> H = LCA([7])
        >>> G + H == LCA([5, 7])  # The `+` operator is overloaded
        True
        >>> G + H == G.direct_sum(H)  # Directs sums two ways
        True
        """
        new_periods = self.periods + other.periods
        new_discrete = self.discrete + other.discrete
        return type(self)(periods=new_periods, discrete=new_discrete)

    def __add__(self, other):
        """
        Override the addition (`+`) operator.
        """
        return self.direct_sum(other)

    def remove_trivial(self):
        """
        Remove trivial groups from the object.

        Returns
        --------
        group : LCA
            The group with trivial groups removed.

        Examples
        ---------
        >>> G = LCA([5, 1, 1])
        >>> G.remove_trivial() == LCA([5])
        True
        """
        def trivial(period, discrete):
            return discrete and (period == 1)

        purged_lists = [(p, d) for (p, d) in self._gen() if not trivial(p, d)]
        new_periods = [p for (p, d) in purged_lists]
        new_discrete = [d for (p, d) in purged_lists]
        return type(self)(periods=new_periods, discrete=new_discrete)

    def rank(self):
        """
        Return the rank of the LCA.

        Returns
        --------
        rank : int
            An integer greater than or equal to 1.

        Examples
        ---------
        >>> G = LCA([5, 6, 1])
        >>> G.rank()
        3
        """
        # TODO: Should trivial groups be removed first?
        return len(self.periods)

    def delete_by_index(self, indices):
        """
        Return a copy with some deleted.

        Parameters
        ----------
        indices

        Returns
        -------

        """
        enu = enumerate
        periods = [p for (i, p) in enu(self.periods) if i not in indices]
        discrete = [p for (i, p) in enu(self.discrete) if i not in indices]
        return type(self)(periods = periods, discrete = discrete)



    @staticmethod
    def _dual_of_single_group(period, discrete):
        """
        Compute the dual of a single group.
        """
        if discrete:
        # Discrete
            if period == 0:
                # Dual of Z is T
                return [1, False]  # Return T
            else:
                # Dual of Z_n is Z_n
                return [period, True]  # Return Z_n

        else:
        # Continuous
            if period == 0:
                # Dual of R is R
                return [0, False]  # Return R
            else:
                # Dual of T is Z
                return [0, True]  # Return Z

    def get_groups(self, slice):
        """
        Return a slice of the LCA.

        Parameters
        ----------
        slice: slice
            A slice object.

        Returns
        -------
        group: LCA
            A slice as specified by the slice object.

        Examples
        ---------
        >>> G = LCA([5, 6, 1])
        >>> G[0:2] == LCA([5, 6])
        True
        >>> G[0] == LCA([5])
        True
        """

        periods = self.periods[slice]
        discrete = self.discrete[slice]
        periods = periods if isinstance(periods, list) else [periods]
        discrete = discrete if isinstance(discrete, list) else [discrete]
        return type(self)(periods = periods, discrete = discrete)

    def __getitem__(self, slice):
        """
        Override the slice (`obj[a:b]`) operator.
        """
        return self.get_groups(slice)

    def __iter__(self):
        """

        Returns
        -------

        """
        if self.is_FGA():
            for period in self.periods:
                yield period

    def __len__(self):
        """
        The rank.

        Returns
        -------

        """
        return len(self.periods)

    def to_list(self):
        """
        Returns periods if FGA.

        Returns
        -------

        """
        if self.is_FGA():
            return self.periods



if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = True)


if __name__ == '__main__':
    G = LCA([0 ,1 ,5 ,0], [True, False, True, False])
    print(G)
    print(G.to_latex())
    print(G.dual())
    print(G.dual().to_latex())
    print(G.dual().remove_trivial())
    print(G.direct_sum(G.dual()))
    print(G.project_element([5, 17, 7, 8.4]))







#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains functions which calculate mapping properties of
free-to-free homomorphisms. All the inputs and outputs are of type
:py:class:`~sympy.matrices.dense.MutableDenseMatrix`.
"""

from sympy import Matrix
from abelian.linalg.factorizations import smith_normal_form
from abelian.linalg.utils import nonzero_diag_as_list


def free_kernel(A):
    """
    Computes the free-to-free kernel of A.

    Let A: Z^n -> Z^m be a free-to-free morphism,
    i.e. a morphism from a free (non-periodic) FGA to
    a free FGA. Associated with the morphism is a kernel
    monomorphism ker(A) such that A * ker(A) * x = 0 for all x.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    ker_A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The kernel monomorphism of A.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0, 1],
    ...             [0, 1, 1]])
    >>> ker_A = free_kernel(A)
    >>> A * ker_A == Matrix([0, 0]) # Check the factorization
    True
    """
    U, S, V = smith_normal_form(A)
    r = len(nonzero_diag_as_list(S))
    return V[:, r:]


def free_cokernel(A):
    """
    Computes the free-to-free cokernel of A.

    Let A: Z^n -> Z^m be a free-to-free morphism,
    i.e. a morphism from a free (non-periodic) FGA to
    a free FGA. Associated with the morphism is a cokernel
    epimorphism coker(A) such that coker(A) * A * x = 0 for all x.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    ker_A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The cokernel epimorphism of A.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> from abelian.linalg.utils import matrix_mod_vector
    >>> A = Matrix([[1, 0],
    ...             [0, 1],
    ...             [1, 1]])
    >>> coker_A = free_cokernel(A)
    >>> quotient = free_quotient(A)
    >>> product = matrix_mod_vector(coker_A * A, quotient)
    >>> product == 0 * product
    True
    """
    U, S, V = smith_normal_form(A, compute_transformation=True)
    return U


def free_image(A):
    """
    Computes the free-to-free image of A.

    Let A: Z^n -> Z^m be a free-to-free morphism,
    i.e. a morphism from a free (non-periodic) FGA to
    a free FGA. Associated with the morphism is an image
    monomorphism im(A) such that im(A) * coim(A) = A.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    ker_A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The image monomorphism of A.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0, 1],
    ...             [0, 1, 1]])
    >>> free_image(A) == Matrix.eye(2)
    True
    >>> free_image(A) * free_coimage(A) == A
    True
    """
    U, S, V = smith_normal_form(A)
    r = len(nonzero_diag_as_list(S))
    return U.inv()[:, :r] * S[:r, :r]


def free_coimage(A):
    """
    Computes the free-to-free coimage of A.

    Let A: Z^n -> Z^m be a free-to-free morphism,
    i.e. a morphism from a free (non-periodic) FGA to
    a free FGA. Associated with the morphism is an coimage
    epimorphism coim(A) such that im(A) * coim(A) = A.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    ker_A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The coimage epimorphism of A.

    Examples
    ---------
    >>> from sympy import Matrix
    >>> A = Matrix([[1, 0],
    ...             [0, 1],
    ...             [1, 1]])
    >>> free_coimage(A) == Matrix.eye(2)
    True
    >>> free_image(A) * free_coimage(A) == A
    True
    """
    U, S, V = smith_normal_form(A)
    r = len(nonzero_diag_as_list(S))
    return V.inv()[:r, :]


def free_quotient(A):
    """
    Compute the quotient group Z^m / im(A).

    Let A: Z^n -> Z^m be a free-to-free morphism,
    i.e. a morphism from a free (non-periodic) FGA to
    a free FGA. Associated with the morphism is an cokernel
    epimorphism coker(A), which maps from Z^m to Z^m / im(A).
    TODO READ UP ON THIS AGAIN.

    Parameters
    ----------
    A : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        A sympy integer matrix.

    Returns
    -------
    quotient : :py:class:`~sympy.matrices.dense.MutableDenseMatrix`
        The quotient of the morphism A.

    Examples
    ---------
    >>> from sympy import diag
    >>> A = diag(1, 2, 3)
    >>> free_quotient(A) == Matrix([1, 1, 6])
    True
    """
    U, S, V = smith_normal_form(A)
    m, n = A.shape
    diagonal = nonzero_diag_as_list(S)
    r = len(diagonal)
    quotient = diagonal + [0] * (m - r)
    return Matrix(quotient)


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose = True)





#!/usr/bin/env python
# -*- coding: utf-8 -*-


class Function:
    """
    A function on a LCA.
    """


    def __init__(self, representation, domain):
        """
        Create a function.

        A function is a ....

        Parameters
        ----------
        representation
        domain
        """
        pass


    def call(self, list_arg, *args, **kwargs):
        """
        Evaluate the function.

        Parameters
        ----------
        arg

        Returns
        -------

        """
        pass

    def pullback(self, morphism):
        """
        Pullback.

        Parameters
        ----------
        morphism

        Returns
        -------

        """
        pass

    def pushfoward(self, morphism):
        """
        Pushfoward.

        Parameters
        ----------
        morphism

        Returns
        -------

        """
        pass

    def compose(self, func):
        """
        Compose with C -> C function.

        Parameters
        ----------
        other

        Returns
        -------

        """
        pass

    def pointwise(self, func, operator):
        """
        Pointwise mult/add/... .

        Parameters
        ----------
        func
        operator

        Returns
        -------

        """

    def convolve(self, other):
        """
        Convolution (if domain is discrete + compact).

        Parameters
        ----------
        other

        Returns
        -------

        """


    def dft(self):
        """
        Discrete fourier transform (if domain is discrete + compact).

        Returns
        -------

        """








#!/usr/bin/env python
# -*- coding: utf-8 -*-
import pkg_resources


__version__ = '0.0.4'


from abelian.linalg.solvers import solve
from abelian.linalg.factorizations import smith_normal_form, hermite_normal_form
from abelian.groups import LCA
from abelian.morphisms import HomFGA, HomLCA, Homomorphism


__all__ = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    Setup file for abelian.

    This file was generated with PyScaffold 2.5.7, a tool that easily
    puts up a scaffold for your new Python project. Learn more under:
    http://pyscaffold.readthedocs.org/
"""

import sys
from setuptools import setup, find_packages
from abelian import __version__
VERSION = __version__

with open('README.rst', 'r') as file:
    README_TEXT = file.read()


def setup_package():
    needs_sphinx = {'build_sphinx', 'upload_docs'}.intersection(sys.argv)
    sphinx = ['sphinx'] if needs_sphinx else []
    setup(setup_requires=['six', 'pyscaffold>=2.5a0,<2.6a0'] + sphinx,
          use_pyscaffold=False,
          name = 'abelian',
          description="Computations on abelian groups.",
          long_description = README_TEXT,
          version=VERSION,
          packages=find_packages(),
          author = "Tommy Odland",
          author_email= 'tommy.odland@student.uib.no',
          url="https://github.com/tommyod/abelian/")


if __name__ == "__main__":
    setup_package()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from random import randint as ri
from sympy import Matrix
from abelian.linalg.factorizations import hermite_normal_form, \
    smith_normal_form


class TestSNF:

    @classmethod
    def setup_class(cls):
        m, n = ri(3, 5), ri(3, 5)
        A = Matrix(m, n, lambda i, j: ri(-9, 9))

        U, S, V = smith_normal_form(A)
        cls.A = A
        cls.U = U
        cls.S = S
        cls.V = V

    def zeros_off_diagonal(self, S):
        """
        Verify that off-diagonals are zero.
        """
        m, n = S.shape
        for i in range(m):
            for j in range(n):
                if i == j:
                    continue
                if S[i, j] != 0:
                    return False
        return True

    def positive_diag(self, S):
        """
        Verifiy that diagonals are positive.
        """
        n = min(S.shape)
        for i in range(n):
            if S[i, i] < 0:
                return False
        return True

    def divisibility_diag(self, S):
        """
        Verify that a_{i}|a_{i+1} for all {i}.
        """
        n = min(S.shape)
        for i in range(n - 1):
            if S[i + 1, i + 1] % S[i, i] != 0:
                return False
        return True

    def test_zeros_off_diagonal(self):
        assert (self.zeros_off_diagonal(self.S))

    def test_positive_diag(self):
        assert (self.positive_diag(self.S))

    def test_divisibility_diag(self):
        assert (self.divisibility_diag(self.S))

    def test_hermite_normal_form(self):
        assert (self.U * self.A * self.V == self.S)

    def test_unimodularity(self):
        assert (self.V.det() in [1, -1]) and (self.U.det() in [1, -1])


class TestHNF:
    @classmethod
    def setup_class(cls):
        """
        Setup state specific to execution of class, which contains tests.
        """
        m, n = ri(3, 5), ri(3, 5)
        A = Matrix(m, n, lambda i, j: ri(-9, 9))

        U, H = hermite_normal_form(A)
        cls.A = A
        cls.U = U
        cls.H = H

    def positive_pivots(self, H):
        """
        Check that all pivots are positive.
        """
        m, n = H.shape
        for j in range(n):
            for i in range(m):
                if H[i, j] == 0:
                    continue
                if H[i, j] > 0:
                    break
                if H[i, j] < 0:
                    return False

        return True

    def left_smaller_than_pivot(self, H):
        """
        Check that all elements k to the left of a pivot h are in the range 0 <= k < h.
        """
        m, n = H.shape
        for j in range(n):
            for i in range(m):
                if H[i, j] > 0:
                    if not all(0 <= H[i, k] < H[i, j] for k in range(0, j)):
                        return False
                    break

        return True

    def test_positive_pivots(self):
        assert self.positive_pivots(self.H)

    def test_smaller_than_pivot(self):
        assert self.left_smaller_than_pivot(self.H)

    def test_hermite_normal_form(self):
        assert self.A * self.U == self.H

    def test_unimodularity(self):
        assert self.U.det() in [1, -1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from random import randint as ri
from sympy import Matrix

from abelian.linalg.free_to_free import free_coimage, free_cokernel, \
    free_image, free_kernel, free_quotient
from abelian.linalg.utils import vector_mod_vector


class TestFreeToFree:
    @classmethod
    def setup_class(cls):
        m, n = ri(3, 5), ri(3, 5)
        A = Matrix(m, n, lambda i, j: ri(-9, 9))
        cls.A = A

        # Compute ker, coker, im, coim and quotient group
        cls.A_ker = free_kernel(A)
        cls.A_coker = free_cokernel(A)
        cls.A_im = free_image(A)
        cls.A_coim = free_coimage(A)
        cls.A_quotient = free_quotient(A)

    def test_im_coim(self):
        """
        Test the image/coimage factorization of A.
        """
        assert (self.A_im * self.A_coim == self.A)

    def test_kernel(self):
        """
        Test that the composition of A and ker(A) is zero.
        """
        sum_entries = sum(abs(entry) for entry in self.A * self.A_ker)
        assert (sum_entries == 0)

    def test_cokernel(self):
        """
        Test that the composition of A and ker(A) is zero in the target group.
        """
        prod = self.A_coker * self.A
        m, n = prod.shape
        ones = Matrix(n, 1, list(1 for i in range(n)))
        prod_in_target_grp = vector_mod_vector(prod * ones, self.A_quotient)
        assert (prod_in_target_grp == self.A_quotient * 0)
